/******************************************************************************
 * Fichier Module : proxy.c - An IGMPv3 Proxy implementation
 ******************************************************************************
 * Fichier    : proxy.c
 * Description: Implementation des divers routines 
 *              Mars 2002
 * Date       : Mars 15, 2000
 * Auteur     : Anis.Ben-Hellel@loria.fr
 * 		sachink@student.utdallas.edu
 * Last Modif : May 13, 2002
 *
 *****************************************************************************/
#include <stdlib.h> 
#include <strings.h> 
#include "igmprt.h"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

vifi_t numvifs;
char buffer[IP_MSFILTER_SIZE(MAX_ADDRS)];
unsigned long upstream;

static char str1[20], str2[20];

void print_source_set(struct in_addr *set_src, int numsrc)
{
	int i;
	printf("Number of sources: %d\n", numsrc);
	for (i = 0; i < numsrc; i++) {
	  printf("%s\n", inet_ntop(AF_INET, &set_src[i].s_addr, str1, 20));
	} 
}

/*
 * set the source list and the source filter
 * on upstream interface
 */

void set_source_filter(
                        igmp_router_t* router,
                        igmp_group_t* gp,
                        unsigned long interface_address,
                        int fmode,
                        int nsources,
                        struct in_addr *sources)

{
       struct ip_msfilter *imsfp;
       int i;
       unsigned int len;

       imsfp = (struct ip_msfilter *) &buffer;
       len = IP_MSFILTER_SIZE(nsources);
       bzero(imsfp, len);

       if (VALID_ADDR(gp->igmpg_addr)) {

         imsfp->imsf_fmode = fmode;
         imsfp->imsf_numsrc = nsources;
#ifdef Linux
	 imsfp->imsf_multiaddr = gp->igmpg_addr.s_addr;
	 imsfp->imsf_interface = interface_address;
#else
         imsfp->imsf_multiaddr.s_addr =gp->igmpg_addr.s_addr;
         imsfp->imsf_interface.s_addr = interface_address;
#endif	 
         for (i=0; i<nsources; i++) {
           imsfp->imsf_slist[i] = sources[i].s_addr;
         }

	 /*use setsockopt*/
	 if (setsockopt(router->igmprt_up_socket, IPPROTO_IP, IP_MSFILTER, 
	 		imsfp, len) < 0 ){
		perror("setsockopt IP_MSFILTER"); 
		print_source_set(sources, nsources);
		printf("Mode: %d Group: %s interface: %s len: %d %d\n", fmode, 
			inet_ntop(AF_INET, &gp->igmpg_addr.s_addr, str1, 20),
			inet_ntop(AF_INET, &interface_address, str2, 20), i,
			len);
	     if (getsockopt(router->igmprt_up_socket, IPPROTO_IP, IP_MSFILTER, 
	 		imsfp, &len) < 0 ){
		perror("getsockopt");
	     }
        	for (i=0; i<imsfp->imsf_numsrc; i++) {
		  printf("src: %s\n",
		  	inet_ntop(AF_INET, &imsfp->imsf_slist[i], str1, 20));
		} 
		printf("Mode: %d Group: %s len: %d\n", imsfp->imsf_fmode, 
			inet_ntop(AF_INET, &imsfp->imsf_multiaddr, str1, 20),
			len);
	 }
       }
       return;
}


/*
 * Open and init the multicast routing in the kernel.
 */

void k_init_proxy(int socket)
{
    int v = 1;
    
    if (setsockopt(socket, IPPROTO_IP, MRT_INIT, (char *)&v, sizeof(int)) < 0)
      perror("setsockopt - MRT_INIT");
}



/*
 * Stops the multicast routing in the kernel.
 */

void k_stop_proxy(int socket)
{
    if (setsockopt(socket, IPPROTO_IP, MRT_DONE, (char *)NULL, 0) < 0)
      perror("setsockopt - MRT_DONE");
}



/* 
 * Add a virtual interface to the kernel 
 * using the pimd API:MRT_ADD_VIF
 * 
 */

int k_proxy_add_vif (int socket,unsigned long vifaddr,vifi_t vifi)
{
	struct vifctl vc;
	int error;
	
	vc.vifc_vifi = vifi;
	vc.vifc_flags = 0;
	vc.vifc_threshold = 0;
	vc.vifc_rate_limit = 0;
	vc.vifc_lcl_addr.s_addr = vifaddr;
	vc.vifc_rmt_addr.s_addr = INADDR_ANY;
	if ((error=setsockopt(socket, IPPROTO_IP, MRT_ADD_VIF,(
		char *)&vc, sizeof(vc)) <0)){
	  perror("setsockopt - MRT_ADD_VIF");
	  return FALSE;
	}
	return TRUE;
}


/*
 * Del an MFC entry from the kernel
 * using pimd API:MRT_DEL_MFC
 */

int k_proxy_del_mfc (int socket, u_long source, u_long group)
{
	struct mfcctl mc;
	
	mc.mfcc_origin.s_addr   = source;
	mc.mfcc_mcastgrp.s_addr = group;
	if (setsockopt(socket, IPPROTO_IP, MRT_DEL_MFC, (char *)&mc, 
			sizeof(mc)) <0){
	  perror("setsockopt - MRT_DEL_MFC");
	  return FALSE;
	}
	return TRUE;
}



/*
 * Install and modify a MFC entry in the kernel (S,G,interface address)
 * using pimd API: MRT_AD_MFC
 */

int k_proxy_chg_mfc(int socket,u_long source, u_long group, vifi_t invif,
		    vifi_t outvif, int fstate)
{
    struct mfcctl mc;
    
    mc.mfcc_origin.s_addr = source;
    mc.mfcc_mcastgrp.s_addr = group;
    
    /* change me to the index of the upstream interface */
    mc.mfcc_parent = invif;

#ifndef Linux    
    mc.mfcc_oif = outvif;
#endif
    mc.mfcc_ttls[outvif] = fstate;
    if (setsockopt(socket, IPPROTO_IP, MRT_ADD_MFC, (char *)&mc, 
    			sizeof(mc)) < 0) {
      perror("setsockopt - MRT_ADD_MFC");
      return(FALSE);
    }
    return(TRUE);
}

/*
 * create entry in the membership database
 */

membership_db*
create_membership(struct in_addr group,int fmode,int numsources,struct in_addr sources[MAX_ADDRS])
{
        membership_db* member;
        int i;
        if ((member = (membership_db*) malloc(sizeof(*member)))) {
                member->membership.group = group;
                member->membership.fmode = fmode;
                member->membership.numsources = numsources;
                for(i=0;i<numsources;i++)
                  member->membership.sources[i].s_addr = sources[i].s_addr;
                member->next = NULL;
                return member;
        }else
                return NULL;
}


/*
 * lookup for a group entry in the membership database
 */

membership_db*
find_membership(membership_db *membership,struct in_addr group)
{
	membership_db* memb;

        for(memb=membership;memb;memb=memb->next) {
          if (memb->membership.group.s_addr == group.s_addr)
            return memb;
	}
        return NULL;
}

/*
 * deleate group entry from membership database
 */

membership_db*
delete_membership(igmp_router_t* igmprt,struct in_addr group)
{
	membership_db *member;
	membership_db *memb;

        assert(igmprt != NULL);
        if((member = find_membership(igmprt->igmprt_membership_db,group)) ==
	    NULL) {
	  return NULL;
	}

        if (igmprt->igmprt_membership_db->membership.group.s_addr != 
		group.s_addr){
          memb = igmprt->igmprt_membership_db;
          while(memb->next->membership.group.s_addr != group.s_addr) {
            memb = memb->next;
	  }
          memb->next = member->next;
          free(member);
        }
        else { /*delete the head*/
          memb = igmprt->igmprt_membership_db;
          igmprt->igmprt_membership_db = memb->next;
          free(memb);
        }
        LOG((LOG_DEBUG, "membership database, group_cleanup: %s\n", 
		inet_ntoa(member->membership.group)));
	return NULL;
}



/*
 * find a source in a in a source list
 */

int find_source(struct in_addr sr,int nsources,struct in_addr *sources)
{
       int i;

       for(i=0;i<nsources;i++)
         if (sources[i].s_addr == sr.s_addr)
           return TRUE;
       return FALSE;
}
/*
 * add multicast group to the membership database
 */

membership_db*
update_multi(igmp_router_t *igmprt,struct in_addr group,int fmode,
		int nsources,struct in_addr sources[MAX_ADDRS])
{

       int i,k;
       membership_db* member;
       struct in_addr sr[MAX_ADDRS];

       /*find corresponding group*/
       if ((member = find_membership(igmprt->igmprt_membership_db,group))) {
         /*update group status using merging rules*/
         member->membership.fmode = (int)member->membership.fmode && (int)fmode;
         if (member->membership.fmode == IGMP_FMODE_INCLUDE) {
           if (fmode == IGMP_FMODE_INCLUDE) {
             for(i=0;i<nsources;i++) {
               if (find_source(sources[i],member->membership.numsources,
	       			member->membership.sources) == FALSE){
                  member->membership.sources[member->membership.numsources].
		  			s_addr = sources[i].s_addr;
                  member->membership.numsources = 
		  		member->membership.numsources + 1;
               }
	     }
           }
           else{
             k = 0;
               for(i=0;i<nsources;i++)
                 if (find_source(sources[i],member->membership.numsources,
		 		 member->membership.sources) == FALSE){
                   sr[k].s_addr = sources[i].s_addr;
                   k = k+1;
                 }
                 member->membership.numsources = k;
                 for(i=0;i<k;i++)
                   member->membership.sources[i].s_addr = sr[i].s_addr;
           }
         }
         else {
           if (fmode == IGMP_FMODE_INCLUDE) {
             k = 0;
             for(i=0;i<member->membership.numsources;i++) {
               if (find_source(member->membership.sources[i],
		 		 nsources,sources) == FALSE){
                 sr[k].s_addr = member->membership.sources[i].s_addr;
                 k = k+1;
               }
               member->membership.numsources = k;
               for(i=0;i<k;i++)
                 member->membership.sources[i].s_addr = sr[i].s_addr;
	     }
           }
           else{
             k = 0;
             for(i=0;i<member->membership.numsources;i++) {
               if (find_source(member->membership.sources[i],
		 		 nsources,sources) == TRUE){
                 sr[k].s_addr = member->membership.sources[i].s_addr;
                 k = k+1;
               }
               member->membership.numsources = k;
               for(i=0;i<k;i++)
                 member->membership.sources[i].s_addr = sr[i].s_addr;
	     }
           }
         }
           //printf("update membership database group: %s, source: %s\n",inet_ntoa(group.s_addr),inet_ntoa(member->membership.sources[0]
       }
       else {
            /*create new entry in the membership database*/
	    LOG((LOG_DEBUG, "New membership entry %s\n", 
	    		inet_ntop(AF_INET, &group, str1, 20)));
            member = create_membership(group,fmode,nsources,sources);
            member->next = igmprt->igmprt_membership_db;
            igmprt->igmprt_membership_db = member;
       }
       return member;
}

void
delete_src_from_membershipdb(igmp_router_t *igmprt,
			     igmp_group_t* group,
			     struct in_addr source)
{
	
	igmp_interface_t *ifp;
	igmp_group_t* gp;
	igmp_src_t *src;
	membership_db* member;
	int i, k;

	/* Check if there are any other interfaces requesting for this (S,G) */
	for (ifp = igmprt->igmprt_interfaces; ifp; ifp = ifp->igmpi_next) {
	  if (ifp->igmpi_addr.s_addr == upstream) {
	    printf("continuing\n");
	    continue;
	  }
	  gp = igmp_interface_group_lookup(ifp, group->igmpg_addr);
	  if (gp) {
	    for (src=gp->igmpg_sources;src;src=(igmp_src_t *)src->igmps_next) {
	      if (src->igmps_source.s_addr == source.s_addr) {
	        printf("returning\n");
	        return;
	      }
	    }
	  }
	}

	/* find corresponding membership and delete the source */
	if ((member = find_membership(igmprt->igmprt_membership_db, 
					group->igmpg_addr))) {
	  k = member->membership.numsources-1;
	  for (i = 0; i < member->membership.numsources; i++) {
	    if (member->membership.sources[i].s_addr == source.s_addr) {
	        member->membership.sources[i].s_addr = 
			        member->membership.sources[k].s_addr;
		member->membership.numsources--;
	        printf("deleting %d %s\n", member->membership.numsources,
	inet_ntop(AF_INET, &member->membership.sources[k].s_addr, str1, 20));
		#if 0
	        set_source_filter(igmprt, group, upstream, 
				  member->membership.fmode,
	  			  member->membership.numsources,
				  member->membership.sources);
	 	#endif
		break;
	    }
	  }
	}
}
